package generated.terminal.copycardMoneySecure.terminal;

//import swt.simulation.SmartCardSimulation;
import swt.terminal.HandleResult;
import swt.util.ByteArray;
import swt.util.HexString;
import de.uniaugsburg.swt.javacard.terminalinterface.SWTReader;
import de.uniaugsburg.swt.javacard.terminalinterface.SWTReaderException;

public class Terminal {
	private Nonce challenge;
	private Secret passphrase;
	private int value;
	private int money;
	private byte state;

	private SWTReader[] reader = new SWTReader[1];
	private Userinterface userinterface;

	private final static int MAX_SPLIT_LENGTH = 50;
	private final static int MAX_APDU_LENGTH = 254;
	private final static int APDU_HEADER_LENGTH = 5;
	private final static int APDU_DATA_LENTH = MAX_APDU_LENGTH
			- APDU_HEADER_LENGTH;

	public Terminal() {
		challenge = new Nonce();
		passphrase = new Secret();
		value = 0;
		money = 0;
		state = State.EXPRESAUTH;

	}

	public Terminal(Nonce challenge, Secret passphrase, int value, int money,
			byte state) {
		this.challenge = challenge;
		this.passphrase = passphrase;
		this.value = value;
		this.money = money;
		this.state = state;

	}

	protected byte[] transmit(byte[] apdu, int port) throws TerminalException,
			SWTReaderException {
		byte[] received = null;
		if (apdu.length <= MAX_APDU_LENGTH) {
			apdu[4] = (byte) (apdu.length - 5);
			received = reader[port].transmit(apdu);
			if (received.length == 2)
				return received;
			received = ByteArray.subarray(received, 0, received.length - 2);
		} else {
			byte[] splittedApdu;
			byte[] apduHeader = ByteArray.subarray(apdu, 0, APDU_HEADER_LENGTH);
			apdu = ByteArray.subarray(apdu, APDU_HEADER_LENGTH);
			int splitCount = apdu.length / APDU_DATA_LENTH;
			if (apdu.length % APDU_DATA_LENTH != 0)
				splitCount++;
			for (int i = 0; i < splitCount; i++) {
				if (i == splitCount - 1) {
					splittedApdu = ByteArray
							.subarray(apdu, i * APDU_DATA_LENTH);
					apduHeader[2] = (byte) 0;
					apduHeader[4] = (byte) splittedApdu.length;
					splittedApdu = ByteArray.append(apduHeader, splittedApdu);
					received = reader[port].transmit(splittedApdu);
					if (received.length == 2)
						return received;
					received = ByteArray.subarray(received, 0,
							received.length - 2);
				} else {
					splittedApdu = ByteArray.subarray(apdu,
							i * APDU_DATA_LENTH, APDU_DATA_LENTH);
					apduHeader[2] = (byte) 1;
					apduHeader[4] = (byte) splittedApdu.length;
					splittedApdu = ByteArray.append(apduHeader, splittedApdu);
					received = reader[port].transmit(splittedApdu);
					if (!(received[0] == (byte) 0x90 && received[1] == (byte) 0x00)) {
						throw new TerminalException(
								"Bad answer from card: part of a long APDU was not answered with OK (0x9000).");
					}
				}
			}
		}

		for (int i = 0; i < MAX_SPLIT_LENGTH; i++) {
			if (received[0] == 0) {
				received = ByteArray.subarray(received, 1);
				return received;
			} else if (received[0] == 1) {
				received = ByteArray.subarray(received, 1);
				byte[] getMore = new byte[]{0, 6, 0, 0, 0, 1, 1};
				byte[] temp_received = reader[port].transmit(getMore);

				temp_received = ByteArray.subarray(temp_received, 0,
						temp_received.length - 2);
				received = ByteArray.append(temp_received[0], received);
				temp_received = ByteArray.subarray(temp_received, 1);
				received = ByteArray.append(received, temp_received);
			} else
				throw new TerminalException(
						"Bad answer from card: answer to a long APDU must begin with 0 or 1.");
		}
		return received;
	}
	private void sendMsg(Message msg) {
		userinterface.send((Userinputs) msg);
	}
	private void sendMsg(Message msg, int port) throws java.lang.Exception {
		Message response;
		switch (port) {
			case Ports.Terminal2Copycard_default_2 :
				response = sendCardMsg(msg, port);
				if (response != null)
					processMessage(response);
				break;

			default :
				stop();
		}
	}
	private Message sendCardMsg(Message msg, int port)
			throws java.lang.Exception {
		byte[] encMsg = new byte[LengthConstants.MAX_ENCODING_LENGTH_MESSAGES];
		short encMsgLen = Coding.getInstance().encode((Codeable) msg, encMsg);
		encMsg = ByteArray.subarray(encMsg, 0, encMsgLen);
		byte[] apdu = ByteArray.append(new byte[]{0, 2, 0, 0, 0}, encMsg);
		byte[] response = transmit(apdu, port);

		if (response.length == 2) {
			if (!(response[0] == (byte) 0x90 && response[1] == (byte) 0x00))
				throw new TerminalException("Cardexception: "
						+ HandleResult.interpretSW(response));
			return null;
		} else {
			Coding.getInstance().decodeInit();
			Codeable cResponse = Coding.getInstance().decode(response);
			return (Message) cResponse;
		}
	}

	public void processMessage(Message inmsg) throws java.lang.Exception {
		switch (inmsg.getCode()) {
			case Code.UREQUESTBALANCE :
				processURequestBalance((URequestBalance) inmsg);
				break;
			case Code.RESREQUESTBALANCE :
				processResRequestBalance((ResRequestBalance) inmsg);
				break;
			case Code.UREQUESTCOPIES :
				processURequestCopies((URequestCopies) inmsg);
				break;
			case Code.RESPAY :
				processResPay((ResPay) inmsg);
				break;
			case Code.UINSERTMONEY :
				processUInsertMoney((UInsertMoney) inmsg);
				break;
			case Code.RESAUTHENTICATE :
				processResAuthenticate((ResAuthenticate) inmsg);
				break;
			default :
				stop();
		}
	}

	public void process(Message msg) throws java.lang.Exception {
		processMessage(msg);
	}

	public void initReader(SWTReader reader, int port) {
		this.reader[port] = reader;
	}

	protected byte[] encode(Codeable c) throws java.lang.Exception {
		byte[] encMsg = new byte[LengthConstants.MAX_ENCODING_LENGTH_MESSAGES];
		short encMsgLen = Coding.getInstance().encode(c, encMsg);
		return ByteArray.subarray(encMsg, 0, encMsgLen);
	}

	public void setUserinterface(Userinterface u) {
		this.userinterface = u;
	}

	protected void stop() throws TerminalException {
		throw new TerminalException("Terminal");
	}

	/**
	 * This method initializes the card with the given parameters that correspond 
	 * to the card attributes. A null value will be ignored by the card.
	 * @throws Exception
	 */
	public void initCard_Copycard(int port, Secret passphrase, byte state)
			throws java.lang.Exception {
		byte cardCode = Code.COPYCARD;
		byte[] apdu = {0, 4, 0, 0, 0, cardCode};
		apdu = ByteArray.append(apdu, (passphrase != null)
				? encode(passphrase)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, new byte[]{Code.BYTE, state});
		transmit(apdu, port);
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processURequestBalance(URequestBalance inmsg)
			throws java.lang.Exception {
		setState(State.IDLE);
		sendMsg(new RequestBalance(), Ports.Terminal2Copycard_default_2);
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processResRequestBalance(ResRequestBalance inmsg)
			throws java.lang.Exception {
		int balance = inmsg.getBalance();

		sendMsg(new UShowBalance(balance));
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processURequestCopies(URequestCopies inmsg)
			throws java.lang.Exception {
		int val = inmsg.getValue();

		if ((val <= 0)) {
			stop();
		} else {
			setValue(val);
			challenge.copy(Nonce.generateNonce());
			setState(State.EXPRESPAY);
			sendMsg(new Pay(value, challenge),
					Ports.Terminal2Copycard_default_2);
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processResPay(ResPay inmsg) throws java.lang.Exception {
		HashedData authhashed = inmsg.getAuthcard();

		if (state == State.EXPRESPAY) {
			setState(State.IDLE);
			if (authhashed.equals(HashedData.hash(new AuthData(Constants.PAY,
					passphrase, challenge, value)))) {
				setMoney((money - value));
				sendMsg(new UIssueCopies(value));
			} else {
				stop();
			}
		} else {
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processUInsertMoney(UInsertMoney inmsg)
			throws java.lang.Exception {
		int val = inmsg.getValue();

		if ((val <= 0)) {
			stop();
		} else {
			setMoney((money + val));
			setValue(val);
			setState(State.EXPRESAUTH);
			sendMsg(new Authenticate(), Ports.Terminal2Copycard_default_2);
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processResAuthenticate(ResAuthenticate inmsg)
			throws java.lang.Exception {
		Nonce challenge = inmsg.getCardletchallenge();

		if (state == State.EXPRESAUTH) {
			setState(State.IDLE);
			AuthData tmp = new AuthData(Constants.LOAD, passphrase, challenge,
					value);
			HashedData hashedauth = HashedData.hash(tmp);
			sendMsg(new Load(value, hashedauth),
					Ports.Terminal2Copycard_default_2);
		} else {
			stop();
		}
	}

	public Nonce getChallenge() {
		return challenge;
	}
	public Secret getPassphrase() {
		return passphrase;
	}
	public int getValue() {
		return value;
	}
	public int getMoney() {
		return money;
	}
	public byte getState() {
		return state;
	}
	public void setChallenge(Nonce challenge) {
		this.challenge = challenge;

	}
	public void setPassphrase(Secret passphrase) {
		this.passphrase = passphrase;

	}
	public void setValue(int value) {
		this.value = value;
	}
	public void setMoney(int money) {
		this.money = money;
	}
	public void setState(byte state) {
		this.state = state;
	}

}
